from django.utils.deconstruct import deconstructible
from jsonschema.validators import Draft202012Validator

from .exceptions import SchemaValidationError


# Validate if json comply with JSONSchema
@deconstructible
class JsonSchemaFieldValidator():

    def __init__(self, schema):
        self.schema = schema

    def __call__(self, value):
        validator = Draft202012Validator(self.schema)
        # Pass `bool` as key function to preserve original order
        error_iterator = sorted(validator.iter_errors(value), key=bool)
        # error_iterator will be empty if no error is found
        if not error_iterator:
            return
        error_list = self.convert_error_list(error_iterator)
        # get nested dictionary of errors message
        raise SchemaValidationError(error_list, error_iterator=error_iterator)

    def convert_error_list(self, error_list):
        # loop through error_list and store jsonschema properties name
        # with json error message as key-value pairs
        result = []
        for error in error_list:
            result.append(f'{error.json_path}: {error.message}')
        return result



from rest_framework.decorators import api_view, permission_classes
from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from tests.serializers import SchemaSerializer


# Create your views here.
@api_view(['GET'])
@permission_classes([AllowAny])
def test_api_view(request):
    json_schema_data = {
        "name":
        1,
        "test_uniqueItems": [1, 2, 3, 4, 4],
        "test_tupleValidation": [24, "Sussex"],
        "test_unevaluatedItems":
        ["life", "universe", "everything", "forty-two"],
        "test_exclusiveMaximum":
        99,
        "test_regex":
        "(888)555-1212 ext. 532",
        "test_integer":
        1,
        "test_additionalProperties":
        "hello",
        "test_minimumProperties": {
            "a": 0
        },
        "children": [{
            "name": "johnny",
            "test_additionalItems": "1"
        }, {
            "children": [{
                "test_uniqueItems": [1, 2, 3],
                "test_additionalItems":
                "1",
                "children": [{
                    "name": "john",
                    "test_uniqueItems": [1, 2, 3],
                    "test_exclusiveMaximum": 99
                }]
            }]
        }]
    }

    serializer = SchemaSerializer(data={"json_data": json_schema_data})
    if serializer.is_valid():
        return Response(serializer.data)
    return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)



from django.test import TestCase
from drf_jsonschema.validators import (Draft202012ValidatorExtended,
                                   JsonSchemaFieldValidator)


# Create your tests here.
class Test(TestCase):
    """
    Test on JsonSchemaFieldValidator
    """

    def test_convert_error_list_path(self):
        """
        Return false if path is not correcly formed
        """
        test_schema = {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string"
                },
                "gender": {
                    "type": "string"
                },
                "children": {
                    "type": "array",
                    "items": {
                        "$ref": "#"
                    }
                }
            },
            "required": ["gender"]
        }
        test_json = {
            "name":
            "Elizabeth",
            "gender":
            "female",
            "children": [{
                "name":
                "Charles",
                "gender":
                "male",
                "children": [{
                    "name":
                    "William",
                    "gender":
                    "male",
                    "children": [{
                        "name": 1,
                        "gender": "female"
                    }, {
                        "name": "Charlotte",
                        "gender": "female"
                    }]
                }, {
                    "name": "Harry",
                    "gender": "male"
                }]
            }]
        }
        validator = Draft202012ValidatorExtended(test_schema)
        error_list = sorted(validator.iter_errors(test_json), key=bool)
        error_list = JsonSchemaFieldValidator.convert_error_list(
            self, error_list)
        self.assertEqual(error_list, [
            '"$.children[0].children[0].children[0].name": 1 is not of type \'string\''
        ])



from django.db import models
import json
from drf_jsonschema.validators import JsonSchemaFieldValidator
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def validate_even(self):
    if 3 % 2 != 0:
        raise ValidationError(
            _('%(value)s is not an even number'),
            params={'value': 3},
        )


def validate_odd(self):
    if 4 % 2 == 0:
        raise ValidationError(
            _('%(value)s is not an odd number'),
            params={'value': 4},
        )


class ModelWithJsonField(models.Model):
    with open('drf_jsonschema/tests/schemas/test_schema.json') as file:
        test_jsonschema = json.load(file)
    json_data = models.JSONField(validators=[
        JsonSchemaFieldValidator(
            schema=test_jsonschema), validate_even, validate_odd
    ])



from rest_framework.exceptions import ValidationError as DRFValidationError
from rest_framework.fields import JSONField
from .exceptions import SchemaValidationError
from .validators import JsonSchemaFieldValidator


class JSONSchemaField(JSONField):

    def __init__(self, *args, **kwargs):
        self.schema = kwargs.pop('schema', None)
        super().__init__(*args, **kwargs)

    def run_validators(self, value):
        error_list = None
        schema_validator = None

        # Find JsonSchemaFieldValidator and removed it
        # This is to prevent DRFValidationError raised before
        # SchemaValidationError is caught.
        # The JsonSchemaFieldValidator will be called and run manually
        # after other validators are run.
        for validator in self.validators:
            if isinstance(validator, JsonSchemaFieldValidator):
                schema_validator = validator
                # Remove JsonSchemaFieldValidator to prevent running twice
                self.validators.remove(validator)

        try:
            # Run other validators except for JsonSchemaFieldValidator
            super().run_validators(value)

        # Update error_dictionary with the error if rest_framework
        # validation_error is raised
        except DRFValidationError as e:
            error_list = e.detail

        try:
            # Instantiate JsonSchemaFieldValidator if the validator
            # is not found
            if not schema_validator:
                # Raise Exception if json schema is not found
                if not self.schema or self.schema is None:
                    raise ValueError("Missing json schema")
                schema_validator = JsonSchemaFieldValidator(self.schema)

            # Run our schema validator manually
            schema_validator(value)

            if error_list:
                raise DRFValidationError(error_list)

        except SchemaValidationError as e:
            # Construct the error dictionary if SchemaValidationError is raised
            error_tree = self.construct_error_dictionary(e.error_iterator)

            if error_list:
                errors = self._get_error_list(error_tree)
                # Append DRFValidation error messages if found
                for error_message in error_list:
                    errors.append(error_message)

            raise DRFValidationError(error_tree)

    # Return a dictionary of errors.
    def construct_error_dictionary(self, error_iterator: dict):
        result = {}
        for error in error_iterator:
            nested = result

            # Append error message if path is empty and error message is string
            if not error.absolute_path and not isinstance(error.message, dict):
                errors = self._get_error_list(nested)
                errors.append(error.message)
                continue
            elif not error.absolute_path and isinstance(error.message, dict):
                errors = self._append_error_message(error.message, nested)
                continue
            last_index = len(error.absolute_path) - 1

            for index, step in enumerate(error.absolute_path):
                # Create nested structure if index not last
                nested = self._get_or_create_nested_dict(nested, step)

                # Append error message if it is last index
                if index == last_index:

                    # Append error message if message is string
                    if not isinstance(error.message, dict):
                        errors = self._get_error_list(nested)
                        errors.append(error.message)
                    # Pass error.message (dict) to the method for structuring
                    # error message
                    else:
                        errors = self._append_error_message(
                            error.message, nested)
        return result

    def _get_or_create_nested_dict(self, obj: dict, key: str):
        value = obj.get(key, {})
        obj[key] = value
        return value

    def _get_error_list(self, obj: dict):
        value = obj.get('errors', [])
        obj['errors'] = value
        return value

    def _append_error_message(self, error: dict, nested: dict):
        # Loop through error.message (dict) to append error message
        for key, error_message in error.items():
            nested = self._get_or_create_nested_dict(nested, key)
            errors = self._get_error_list(nested)
            errors.append(error_message[0])
        return errors



from django.core.exceptions import ValidationError


# Custom class to validate schema
class SchemaValidationError(ValidationError):

    def __init__(self, message, code=None, params=None, error_iterator=None):
        super().__init__(message, code, params)
        self.error_iterator = error_iterator



from drf_jsonschema.serializers import JsonSchemaModelSerializer
from tests.models import ModelWithJsonField


class SchemaSerializer(JsonSchemaModelSerializer):

    class Meta:
        model = ModelWithJsonField
        fields = '__all__'



from rest_framework import serializers
from drf_jsonschema.fields import JSONSchemaField
from django.db import models


class JsonSchemaModelSerializer(serializers.ModelSerializer):
    serializer_field_mapping = serializers.ModelSerializer\
        .serializer_field_mapping.copy()
    serializer_field_mapping[models.JSONField] = JSONSchemaField



from jsonschema.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def format_message(message):
    """
    Format error message to keep a consistent outcome
    """
    return [_(message)]


def required(validator, required, instance, schema):
    """
    Error for required field not provided
    """
    if not validator.is_type(instance, "object"):
        return
    for property in required:
        if property not in instance:
            yield ValidationError({
                property:
                format_message("This field is required.", ),
            })


def dependentRequired(validator, dependentRequired, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    for property, dependency in dependentRequired.items():
        if property not in instance:
            continue

        for each in dependency:
            if each not in instance:
                message = f"{each!r} is a dependency of {property!r}"
                yield ValidationError({
                    each: format_message(message),
                })